@@ -25,8 +25,10 @@ from jbutils.utils.utils import (
 from jbutils.utils.config import Configurator, get_default_cfg_files
 from jbutils.utils import config as jbcfg
 from jbutils.utils import utils as jbutils
+from jbutils.utils import cli_util
 
 __all__ = [
+    "cli_util",
     "Configurator",
     "Consts",
     "copy_to_clipboard",
