@@ -20,6 +20,11 @@ from collections.abc import Callable
 from dataclasses import dataclass
 from typing import Any, Optional
 
+import jsonlines
+
+from jsonlines.jsonlines import JSONValue
+from PIL import Image, ImageFile
+from PIL.ImageFile import ImageFile
 from ruamel.yaml import YAML
 from ruamel.yaml.comments import CommentedMap, CommentedSeq
 from ruamel.yaml.tokens import CommentToken
@@ -32,6 +37,26 @@ yaml = YAML()
 yaml.indent = 2
 
 Predicate = Callable[[T], bool]
+IMAGE_EXTS = [
+    ".png",
+    ".jpg",
+    ".jpeg",
+    ".gif",
+    ".bmp",
+    ".tiff",
+    ".webp",
+    ".avif",
+    ".svg",
+    ".ico",
+]
+
+SUPPROTED_EXTS = [
+    "csv",
+    "json",
+    "jsonl",
+    "yaml",
+    "yml",
+] + IMAGE_EXTS
 
 
 class Consts:
@@ -63,10 +88,12 @@ def read_file(
     path: str,
     mode: str = "r",
     encoding: str | None = Consts.encoding,
-    as_lines: bool = False,
     default_val: Any = None,
+    as_lines: bool = False,
     as_dicts: bool = False,
-) -> str | dict | list:
+    *args,
+    **kwargs,
+) -> str | dict | list | ImageFile:
     """Read data from a file
 
     Args:
@@ -74,11 +101,17 @@ def read_file(
         mode (str, optional): IO mode to use. Defaults to "r".
         encoding (str, optional): Encoding format to use.
             Defaults to "latin-1".
+        default_val (Any, optional): Value to return if the file is not found.
+            Defaults to None.
         as_lines (bool, optional): If reading a regular text file,
             True will return the value of readlines() instead of read().
             Defaults to False.
-        default_val (Any, optional): Value to return if the file is not found.
-            Defaults to None.
+        as_dicts (bool, options): If reading a csv file, return lines as
+            dicts if True
+        *args (args, optional): Special args to pass to the associated
+            reader class
+        **kwargs (kwargs, optional): Special kwargs to pass to the associated
+            reader class
 
     Returns:
         str | dict | list: The data read from the file. If the file is
@@ -91,14 +124,16 @@ def read_file(
         print(f"Warning: Path '{path}' does not exist")
         return default_val
 
-    _, ext = os.path.splitext(path)
+    ext = get_ext(path).lower()
 
-    with open(path, mode, encoding=encoding) as fs:
-        match ext.lower():
+    with open(path, mode, encoding=encoding, *args, **kwargs) as fs:
+        match ext:
             case ".yaml" | ".yml":
                 return yaml.load(stream=fs)
             case ".json":
                 return json.load(fs)
+            case ".jsonl":
+                return [item for item in jsonlines.Reader(fs).iter()]
             case ".csv":
                 data = list(csv.reader(fs))
                 if as_dicts:
@@ -113,18 +148,41 @@ def read_file(
                     return [dict(zip(cols, vals)) for vals in data]
                 return data
             case _:
+                if ext.lower() in IMAGE_EXTS:
+                    return Image.open(fs)
                 if as_lines:
                     return fs.readlines()
                 else:
                     return fs.read()
 
 
+def to_jsonl(items: list, check_serial: bool = True) -> list:
+    new_items = []
+    for item in items:
+        if isinstance(item, str):
+            # If it's a str, just try to convert to a json serialized
+            # structure, since a raw str isn't a valid jsonl line value
+            new_items.append(json.loads(item))
+        else:
+            # Otherwise, alternate dumps and loads to make sure item is
+            # serializable if check_serial is true, otherwise just use the raw value
+            value = json.loads(json.dumps(item)) if check_serial else item
+            new_items.append(value)
+    return new_items
+
+
+def ext_supported(ext: str, data: Any = None) -> bool:
+    return ext in SUPPROTED_EXTS or isinstance(data, ImageFile)
+
+
 def write_file(
     path: str,
     data: Any,
     mode: str = "w",
     encoding: str | None = Consts.encoding,
     indent: int = 4,
+    *args,
+    **kwargs,
 ) -> None:
     """Write text to a file
 
@@ -134,19 +192,33 @@ def write_file(
         mode (str, optional): Read/write mode. Defaults to "w".
         encoding (str, optional): Encoding to write with. Defaults to ENCODING.
         indent (int, optional): Indent to apply if JSON. Defaults to 4.
-        sort_keys (bool, optional): Whether to sort the output keys for YAML.
-            Defaults to False.
+        *args (args, optional): Special args to pass to the associated
+            writer class
+        **kwargs (kwargs, optional): Special kwargs to pass to the associated
+            writer class
     """
 
-    _, ext = os.path.splitext(path)
+    ext = get_ext(path).lower()
 
-    with open(path, mode, encoding=encoding) as fs:
-        match ext.lower():
+    wr_args = args if not ext_supported(ext, data) else ()
+    wr_kwargs = kwargs if not ext_supported(ext, data) else {}
+
+    with open(path, mode, encoding=encoding, *wr_args, **wr_kwargs) as fs:
+        match ext:
             case ".yml" | ".yaml":
-                yaml.dump(data, fs)
+                yaml.dump(data, fs, *args, **kwargs)
             case ".json":
-                json.dump(data, fs, indent=indent)
+                json.dump(data, fs, indent=indent, *args, **kwargs)
+            case ".jsonl":
+                writer = jsonlines.Writer(fs, *args, **kwargs)
+                if isinstance(data, list):
+                    lines = to_jsonl(data)
+                else:
+                    lines = to_jsonl([data])
+                writer.write_all(lines)
             case _:
+                if isinstance(data, ImageFile):
+                    data.save(fs, *args, **kwargs)
                 if isinstance(data, list):
                     fs.writelines(data)
                 else:
@@ -920,8 +992,9 @@ def to_csv_line(items: list[str]) -> str:
 def add_common_args(
     parser: argparse.ArgumentParser,
     proj_path: str,
+    proj_dir: str = "",
     overrides: dict[str, CommandArg] | None = None,
-) -> Callable[..., None]:
+) -> Callable[..., argparse.Namespace]:
     overrides = overrides or {}
 
     cmd_args: list[CommandArg] = [
@@ -950,17 +1023,20 @@ def add_common_args(
         parser.add_argument(*arg.name_or_flags, action=arg.action, help=arg.help)
 
     proj_path = os.path.abspath(proj_path)
+    proj_dir = proj_dir or os.path.dirname(os.path.dirname(proj_path))
 
-    def handle_common() -> None:
+    def handle_common() -> argparse.Namespace:
         code, cd = cmd_args
-        args = vars(parser.parse_args())
-        if args.get(code.arg_name):
+        args = parser.parse_args()
+        arg_props = vars(args)
+        if arg_props.get(code.arg_name):
             cmdx(f"code {proj_path}")
             sys.exit(0)
-        if args.get(cd.arg_name):
-            path = os.path.dirname(os.path.dirname(proj_path))
+        if arg_props.get(cd.arg_name):
+            path = proj_dir
             copy_to_clipboard(f"cd {path}")
             sys.exit(0)
+        return args
 
     return handle_common
 
